import aframexr
import math
import unittest

from bs4 import BeautifulSoup

from aframexr.api.filters import FilterTransform
from tests.constants import *  # Constants used for testing


def _all_theta_sum_is_360_degrees(pie_chart: aframexr.Chart) -> bool:
    """Verify that the sum for the theta length of every slice is 360 degrees."""

    total_theta_length = 0

    soup = BeautifulSoup(pie_chart.to_html(), 'html.parser')
    slices = soup.find_all('a-cylinder')
    for s in slices:
        total_theta_length += float(s['theta-length'])
    if not math.isclose(total_theta_length, 360):
        return False
    return True

def _slices_are_well_placed(pie_chart: aframexr.Chart) -> bool:
    """Verify that the slices are well-placed in the pie chart (relative position has to be '0 0 0' for all)."""

    soup = BeautifulSoup(pie_chart.to_html(), 'html.parser')
    slices = soup.find_all('a-cylinder')
    for s in slices:
        if s['position'] != '0 0 0':
            return False
    return True


class TestMarkArcOK(unittest.TestCase):
    """Pie chart OK tests."""

    def test_simple(self):
        """Simple pie chart creation."""

        pie_chart = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales')
        pie_chart.show()
        assert _all_theta_sum_is_360_degrees(pie_chart)
        assert _slices_are_well_placed(pie_chart)

    def test_data_format(self):
        """Pie chart changing data format creation."""

        for d in DATA_FORMATS:
            pie_chart = aframexr.Chart(d).mark_arc().encode(color='model', theta='sales')
            pie_chart.show()
            assert _all_theta_sum_is_360_degrees(pie_chart)
            assert _slices_are_well_placed(pie_chart)

    def test_position(self):
        """Pie chart changing position creation."""

        for p in POSITIONS:
            pie_chart = aframexr.Chart(DATA, position=p).mark_arc().encode(color='model', theta='sales')
            pie_chart.show()
            assert _all_theta_sum_is_360_degrees(pie_chart)
            assert _slices_are_well_placed(pie_chart)

    def test_position_format(self):
        """Pie chart changing position format creation."""

        for p in POSITION_FORMATS:
            pie_chart = aframexr.Chart(DATA, position=p).mark_arc().encode(color='model', theta='sales')
            pie_chart.show()
            assert _all_theta_sum_is_360_degrees(pie_chart)
            assert _slices_are_well_placed(pie_chart)

    def test_rotation(self):
        """Pie chart changing rotation creation."""

        for r in ROTATIONS:
            pie_chart = aframexr.Chart(DATA, rotation=r).mark_arc().encode(color='model', theta='sales')
            pie_chart.show()
            assert _all_theta_sum_is_360_degrees(pie_chart)
            assert _slices_are_well_placed(pie_chart)

    def test_rotation_format(self):
        """Pie chart changing rotation format creation."""

        for r in ROTATION_FORMATS:
            pie_chart = aframexr.Chart(DATA, rotation=r).mark_arc().encode(color='model', theta='sales')
            pie_chart.show()
            assert _all_theta_sum_is_360_degrees(pie_chart)
            assert _slices_are_well_placed(pie_chart)

    def test_radius(self):
        """Pie chart changing radius creation."""

        for r in MARK_ARC_RADIUS:
            pie_chart = aframexr.Chart(DATA).mark_arc(radius=r).encode(color='model', theta='sales')
            pie_chart.show()
            assert _all_theta_sum_is_360_degrees(pie_chart)
            assert _slices_are_well_placed(pie_chart)

    def test_filter(self):
        """Pie chart changing filter creation."""

        for eq in FILTER_EQUATIONS:
            for f in [eq, FilterTransform.from_string(eq)]:  # Filter using equation and using FilterTransform object
                pie_chart = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales').transform_filter(f)
                pie_chart.show()
                assert _all_theta_sum_is_360_degrees(pie_chart)
                assert _slices_are_well_placed(pie_chart)

    def test_position_rotation_radius_filter(self):
        """Pie chart changing position, rotation, radius and filter creation."""

        for pos, rot, rad, fil in zip(POSITIONS, ROTATIONS, MARK_ARC_RADIUS, FILTER_EQUATIONS):
            pie_chart = (aframexr.Chart(DATA, position=pos, rotation=rot).mark_arc(radius=rad)
                         .encode(color='model',theta='sales').transform_filter(fil))
            pie_chart.show()
            assert _all_theta_sum_is_360_degrees(pie_chart)
            assert _slices_are_well_placed(pie_chart)


class TestMarkArcError(unittest.TestCase):
    """Pie chart error tests."""

    def test_position_error(self):
        """Pie chart position error."""

        for p in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_arc().encode(color='model', theta='sales')
            assert str(error.exception) == f'The position: {p} is not correct. Must be "x y z".'

        for p in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_arc().encode(color='model', theta='sales')
            assert str(error.exception) == 'The position values must be numeric.'

    def test_rotation_error(self):
        """Pie chart rotation error."""

        for r in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_arc().encode(color='model', theta='sales')
            assert str(error.exception) == f'The rotation: {r} is not correct. Must be "x y z".'

        for r in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_arc().encode(color='model', theta='sales')
            assert str(error.exception) == 'The rotation values must be numeric.'

    def test_radius_error(self):
        """Pie chart radius error."""

        for r in NOT_GREATER_THAN_0_MARK_ARC_RADIUS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_arc(radius=r).encode(color='model', theta='sales')
            assert str(error.exception) == 'The radius must be greater than 0.'

    def test_encoding_error(self):
        """Pie chart encoding error."""

        for e in NON_EXISTING_MARK_ARC_ENCODINGS:
            with self.assertRaises(KeyError) as error:
                pie_chart = aframexr.Chart(DATA).mark_arc().encode(**e)
                pie_chart.show()
            assert 'Data has no field ' in str(error.exception)

        for e in NOT_VALID_MARK_ARC_ENCODINGS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_arc().encode(**e)
            assert str(error.exception) in ['Parameter theta must be specified in arc chart.',
                                            'Parameter color must be specified in arc chart.']

    def test_filter_warning(self):
        """Pie chart filter warning."""

        for f in WARNING_FILTER_EQUATIONS:
            with self.assertWarns(UserWarning) as warning:
                filt_chart = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales').transform_filter(f)
                filt_chart.show()
            assert str(warning.warning) == f'Data does not contain values for the filter: {f}.'

    def test_filter_error(self):
        """Pie chart filter error."""

        for f in ERROR_FILTER_EQUATIONS:
            with self.assertRaises(SyntaxError) as error:
                filt_chart = aframexr.Chart(DATA).mark_arc().encode(color='model', theta='sales').transform_filter(f)
                filt_chart.show()
            assert str(error.exception) in ['Incorrect syntax, must be datum.{field} == {value}',
                                            'Incorrect syntax, must be datum.{field} > {value}',
                                            'Incorrect syntax, must be datum.{field} < {value}']



"""AframeXR entity creator"""

import copy
import io
import numpy as np
import os
import pandas as pd
import urllib.request, urllib.error
import warnings

from itertools import cycle, islice
from pandas import DataFrame, Series

from aframexr.utils.constants import *


AXIS_DICT_TEMPLATE = {
    'x': {'start': None, 'end': None, 'labels_pos': [], 'labels_values': [], 'labels_rotation': ''},
    'y': {'start': None, 'end': None, 'labels_pos': [], 'labels_values': [], 'labels_rotation': ''},
    'z': {'start': None, 'end': None, 'labels_pos': [], 'labels_values': [], 'labels_rotation': ''}
}
"""Axis dictionary template for chart creation."""

GROUP_DICT_TEMPLATE = {'pos': '', 'rotation': ''}
"""Group dictionary template for group base specifications creation."""


def _get_data_from_url(url: str) -> DataFrame:
    """Loads the data from the URL (could be a local path) and returns it as a DataFrame."""

    if url.startswith(('http://', 'https://')):  # Data is stored in a URL
        try:
            with urllib.request.urlopen(url) as response:
                file_type = response.info().get_content_type()
                data = io.BytesIO(response.read())  # For pandas
        except urllib.error.URLError:
            raise IOError(f'Could not load data from URL: {url}.')
    else:  # Data is stored in a local file
        path = os.path.normpath(url)
        if not os.path.exists(path):
            raise FileNotFoundError(f'Local file "{path}" was not found.')

        data = open(path, 'rb')
        _, file_type = os.path.splitext(path)
        file_type = file_type.lower()
    try:
        if 'csv' in file_type:  # Data is in CSV format
            df_data = pd.read_csv(data)
        elif 'json' in file_type:
            df_data = pd.read_json(data)
        else:
            raise NotImplementedError(f'Unsupported file type: {file_type}.')
    except Exception as e:
        raise IOError(f'Error when processing data. Error: {e}.')

    if data and not url.startswith(('http', 'https')):
        data.close()  # Close the file

    return df_data


def _get_raw_data(chart_specs: dict) -> DataFrame:
    """Returns the raw data from the chart specifications, transformed if necessary."""

    # Get the raw data of the chart
    data_field = chart_specs['data']
    if data_field.get('url'):  # Data is stored in a file
        raw_data = _get_data_from_url(data_field['url'])
    elif data_field.get('values'):  # Data is stored as the raw data
        json_data = data_field['values']
        raw_data = pd.DataFrame(json_data)
    else:
        raise ValueError('Data specifications has no correct syntaxis, must have field "url" or "values".')

    # Transform data (if necessary)
    from aframexr.api.aggregate import AggregatedFieldDef  # To avoid circular import error
    from aframexr.api.filters import FilterTransform
    transform_field = chart_specs.get('transform')
    if transform_field:

        for filter_transformation in transform_field:  # The first transformations are the filters
            if filter_transformation.get('filter'):
                filter_object = FilterTransform.from_string(filter_transformation['filter'])

                if isinstance(filter_object.value, str):
                    filter_object.value = f'"{filter_object.value}"'  # Add quotes to the value for query search
                raw_data = filter_object.get_filtered_data(raw_data)
                if raw_data.empty:  # Data does not contain any value for the filter
                    warnings.warn(f'Data does not contain values for the filter: {filter_transformation["filter"]}.')

        for non_filter_transf in transform_field:  # Non-filter transformations
            groupby = set(non_filter_transf.get('groupby')) if non_filter_transf.get('groupby') else set()
            if non_filter_transf.get('aggregate'):

                for aggregate in non_filter_transf.get('aggregate'):
                    aggregate_object = AggregatedFieldDef.from_dict(aggregate)

                    encoding_channels = {  # Using a set to have the possibility of getting differences
                        ch_spec['field'] for ch_spec in chart_specs['encoding'].values()  # Take the encoding channels
                        if ch_spec['field'] != aggregate_object.as_field  # Except the aggregate field channel
                    }

                    if groupby:
                        not_defined_channels = encoding_channels - set(groupby)  # Difference between sets
                        if not_defined_channels:  # There are channels in encoding_channels not defined in groupby
                            raise ValueError(
                                f'Encoding channel(s) "{not_defined_channels}" must be defined in aggregate groupby: '
                                f'{groupby}, otherwise that fields will disappear.'
                            )
                    else:
                        groupby = list(encoding_channels)  # Use the encoding channels as groupby
                    raw_data = aggregate_object.get_aggregated_data(raw_data, groupby)

    # Aggregate in encoding
    encoding_channels = chart_specs['encoding']
    aggregate_fields = [ch['field'] for ch in encoding_channels.values() if ch.get('aggregate')]
    aggregate_ops = [ch['aggregate'] for ch in encoding_channels.values() if ch.get('aggregate')]
    groupby_fields = [spec['field'] for spec in encoding_channels.values() if not spec.get('aggregate')]

    for ag in range(len(aggregate_fields)):
        aggregate_object = AggregatedFieldDef(aggregate_ops[ag], aggregate_fields[ag])
        raw_data = aggregate_object.get_aggregated_data(raw_data, groupby_fields)

    return raw_data


class ChartCreator:
    """Chart creator base class"""

    def __init__(self, chart_specs: dict):
        base_position = chart_specs.get('position', DEFAULT_CHART_POS)
        [self._base_x, self._base_y, self._base_z] = [float(pos) for pos in base_position.split()]  # Base position
        self._encoding = chart_specs.get('encoding')  # Encoding and parameters of the chart
        rotation = chart_specs.get('rotation', DEFAULT_CHART_ROTATION)  # Rotation of the chart
        [self._x_rotation, self._y_rotation, self._z_rotation] = [float(rot) for rot in rotation.split()]

    @staticmethod
    def create_object(chart_type: str, chart_specs: dict):
        """Returns a ChartCreator instance of the specific chart type."""

        CREATOR_MAP = {
            'arc': ArcChartCreator,
            'bar': BarChartCreator,
            'gltf': GLTFModelCreator,
            'image': ImageCreator,
            'point': PointChartCreator,
        }

        if chart_type not in CREATOR_MAP:
            raise ValueError(f'Invalid chart type: {chart_type}.')
        return CREATOR_MAP[chart_type](chart_specs)

    def get_group_specs(self) -> dict:
        """Returns a dictionary with the base specifications for the group of elements."""

        group_specs = copy.deepcopy(GROUP_DICT_TEMPLATE)
        group_specs.update({'pos': f'{self._base_x} {self._base_y} {self._base_z}',
                            'rotation': f'{self._x_rotation} {self._y_rotation} {self._z_rotation}'})
        return group_specs


class ArcChartCreator(ChartCreator):
    """Arc chart creator class."""

    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._raw_data = _get_raw_data(chart_specs)  # Raw data
        self._radius = chart_specs['mark'].get('radius', DEFAULT_PIE_RADIUS)  # Radius
        self._set_rotation()
        self._color_data = Series()
        self._theta_data = Series()

    def _set_rotation(self):
        """Sets the rotation of the pie chart."""

        pie_rotation = DEFAULT_PIE_ROTATION.split()  # Default rotation for the pie chart to look at the camera
        self._x_rotation = self._x_rotation + float(pie_rotation[0])
        self._y_rotation = self._y_rotation + float(pie_rotation[1])
        self._z_rotation = self._z_rotation + float(pie_rotation[2])

    def _set_elements_theta(self) -> tuple[Series, Series]:
        """Returns a tuple with a Series storing the theta start of each element, and another storing theta length."""

        sum_data = self._theta_data.sum()  # Sum all the values
        theta_length = (self._theta_data / sum_data) * 360  # Series of theta lengths (in degrees)
        theta_start = theta_length.cumsum().shift(1, fill_value=0)  # Accumulative sum (first value is 0)
        return theta_start, theta_length

    def _set_elements_colors(self) -> Series:
        """Returns a Series of the color for each element composing the chart."""

        colors = cycle(AVAILABLE_COLORS)  # Color cycle iterator
        element_colors = Series(islice(colors, len(self._color_data)))  # Take len(self._color_data) colors
        return element_colors

    def get_elements_specs(self) -> list[dict]:
        """Returns a list of dictionaries with the specifications for each element of the chart."""

        if self._raw_data.empty:  # There is no data to display
            return []

        # Axis
        x_coordinates = Series(data=np.full(len(self._raw_data), 0), index=self._raw_data.index)
        y_coordinates = Series(data=np.full(len(self._raw_data), 0), index=self._raw_data.index)
        z_coordinates = Series(data=np.full(len(self._raw_data), 0), index=self._raw_data.index)

        # Radius
        radius = Series(
            data=np.full(len(self._raw_data), self._radius),
            index=self._raw_data.index
        )

        # Theta
        theta_field = self._encoding['theta']['field']
        try:
            self._theta_data = self._raw_data[theta_field]
        except KeyError:
            raise KeyError(f'Data has no field "{theta_field}".')
        theta_starts, theta_lengths = self._set_elements_theta()

        # Color
        color_field = self._encoding['color']['field']
        try:
            self._color_data = self._raw_data[color_field]
        except KeyError:
            raise KeyError(f'Data has no field "{color_field}".')
        colors = self._set_elements_colors()

        # Id
        ids_series = [self._color_data.astype(str), self._theta_data.astype(str)]
        ids = ids_series[0].str.cat(others=ids_series[1:], sep=' : ', na_rep='?')  # Concatenate values of the series

        # Return values
        pos_series = x_coordinates.astype(str).str.cat([y_coordinates.astype(str), z_coordinates.astype(str)], sep=' ')

        temp_df = pd.DataFrame({
            'id': ids,
            'pos': pos_series,
            'radius': radius,
            'theta_start': theta_starts,
            'theta_length': theta_lengths,
            'color': colors
        })
        elements_specs = temp_df.to_dict(orient='records')  # Transform DataFrame into a list of dictionaries
        return elements_specs

    def get_axis_specs(self) -> dict:
        """Returns a dictionary with the specifications for each axis of the chart."""

        axis_specs = copy.deepcopy(AXIS_DICT_TEMPLATE)
        return axis_specs  # Arc chart have no axis


class BarChartCreator(ChartCreator):
    """Bar chart creator class."""

    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._raw_data = _get_raw_data(chart_specs)  # Raw data
        self._bar_width = chart_specs['mark'].get('width', DEFAULT_BAR_WIDTH)  # Width of the bar
        self._max_height = chart_specs.get('height', DEFAULT_MAX_HEIGHT)  # Maximum height of the bar chart
        self._x_data: Series | None = None
        self._y_data: Series | None = None
        self._z_data: Series | None = None

    def _set_bars_colors(self) -> Series:
        """Returns a Series of the color for each bar composing the bar chart."""

        colors = cycle(AVAILABLE_COLORS)  # Color cycle iterator
        bars_colors = Series(islice(colors, len(self._raw_data)))  # Take len(self._raw_data) colors from the cycle
        return bars_colors

    def _set_bars_heights(self) -> Series:
        """Returns a Series of the height for each bar composing the bar chart."""

        if self._y_data is None:
            heights = Series(  # Series of DEFAULT_BAR_HEIGHT_WHEN_NO_Y_AXIS values
                data=np.full(len(self._raw_data), DEFAULT_BAR_HEIGHT_WHEN_NO_Y_AXIS),
                index=self._raw_data.index
            )
        else:
            max_value = self._y_data.max()
            heights = (self._y_data / max_value) * self._max_height
        return heights

    def _set_x_coordinates(self) -> Series:
        """Returns a Series of the x coordinates for each bar composing the bar chart."""

        relative_x_start = self._bar_width / 2  # Shift because of box creations

        if self._x_data is None:  # No field for x-axis
            x_coordinates = Series(  # Series of relative_x_start values
                data=np.full(len(self._raw_data), relative_x_start),
                index=self._raw_data.index
            )
        else:  # Field for x-axis
            x_coordinates = Series(
                data=relative_x_start + (np.arange(len(self._raw_data)) * self._bar_width),
                index=self._raw_data.index
            )
        return x_coordinates

    def _set_z_coordinates(self) -> Series:
        """Returns a Series of the z coordinates for each bar composing the bar chart."""

        relative_z_start = - DEFAULT_BAR_DEPTH / 2  # Shift because of box creations

        if self._z_data is None:
            z_coordinates = Series(  # Series of relative_z_start values
                data=np.full(len(self._raw_data), relative_z_start),
                index=self._raw_data.index
            )
        else:
            categories = pd.Categorical(self._z_data)  # Assign one number for each category (faster indexation)
            z_coordinates_map = np.linspace(  # Array of equally spaced values
                start=relative_z_start,  # First value of the array
                stop=-DEFAULT_MAX_DEPTH + (DEFAULT_BAR_DEPTH / 2),  # Last value of the array
                num=len(categories.categories)  # Number of elements
            )
            z_coordinates = pd.Series(z_coordinates_map[categories.codes], index=self._z_data.index)
        return z_coordinates

    def get_elements_specs(self) -> list[dict]:
        """Returns a list of dictionaries with the specifications for each element of the chart."""

        if self._raw_data.empty:  # There is no data to display
            return []

        # X-axis
        if self._encoding.get('x'):
            x_field = self._encoding['x']['field']  # Field of the x-axis
            try:
                self._x_data = self._raw_data[x_field]
            except KeyError:
                raise KeyError(f'Data has no field "{x_field}".')

        # Y-axis
        if self._encoding.get('y'):
            y_field = self._encoding['y']['field']  # Field of the y-axis
            try:
                self._y_data = self._raw_data[y_field]
            except KeyError:
                raise KeyError(f'Data has no field "{y_field}".')

        # Z-axis
        if self._encoding.get('z'):
            z_field = self._encoding['z']['field']  # Field of the z-axis
            try:
                self._z_data = self._raw_data[z_field]
            except KeyError:
                raise KeyError(f'Data has no field "{z_field}".')

        bar_widths = Series(  # Series of self._bar_width values
            data=np.full(len(self._raw_data), self._bar_width),
            index=self._raw_data.index
        )
        x_coordinates = self._set_x_coordinates()  # X-axis coordinate for each bar

        bar_heights = self._set_bars_heights()  # Series of the height for each bar
        y_coordinates = bar_heights / 2  # Y-axis coordinates is the height of the bar / 2 (because of box creation)

        z_coordinates = self._set_z_coordinates()

        # Color
        colors = self._set_bars_colors()

        # Id
        ids_series = []
        if self._x_data is not None:
            ids_series.append(self._x_data.astype(str))
        if self._y_data is not None:
            ids_series.append(self._y_data.astype(str))
        if self._z_data is not None:
            ids_series.append(self._z_data.astype(str))

        ids = ids_series[0].str.cat(others=ids_series[1:], sep=' : ', na_rep='?')  # Concatenate values of the series

        # Return values
        pos_series = x_coordinates.astype(str).str.cat([y_coordinates.astype(str), z_coordinates.astype(str)], sep=' ')
        temp_df = pd.DataFrame({
            'id': ids,
            'pos': pos_series,
            'width': bar_widths,
            'height': bar_heights,
            'color': colors,
        })
        elements_specs = temp_df.to_dict(orient='records')  # Transform DataFrame into a list of dictionaries
        return elements_specs

    def get_axis_specs(self) -> dict:
        """Returns a dictionary with the specifications for each axis of the chart."""

        axis_specs = copy.deepcopy(AXIS_DICT_TEMPLATE)

        if self._raw_data.empty:  # There is no data to display
            return axis_specs

        # ---- X-axis ----
        # Axis line
        display_axis = True
        try:
            display_axis = self._encoding['x']['axis'] if self._encoding.get('x') else False
        except KeyError or display_axis is True:  # Display axis if key 'axis' not found (default display axis) or True
            axis_specs['x']['start'] = '0 0 0'
            axis_specs['x']['end'] = f'{self._bar_width * len(self._x_data)} 0 0'

            # Axis labels
            x_coords = self._set_x_coordinates()  # X-axis value for each bar
            y_coords = Series(  # Series of LABELS_Y_DELTA values
                data=np.full(len(self._x_data), LABELS_Y_DELTA),
                index=self._x_data.index
            )
            z_coords = Series(  # Series of LABELS_Z_DELTA values
                data=np.full(len(self._x_data), LABELS_Z_DELTA),
                index=self._x_data.index
            )
            label_pos_series = x_coords.astype(str).str.cat([y_coords.astype(str), z_coords.astype(str)], sep=' ')

            axis_specs['x']['labels_pos'] = label_pos_series.tolist()
            axis_specs['x']['labels_values'] = self._x_data.tolist()
            axis_specs['x']['labels_rotation'] = '-90 0 -90'

        # ---- Y-axis ----
        # Axis line
        display_axis = True
        try:
            display_axis = self._encoding['y']['axis'] if self._encoding.get('y') else False
        except KeyError or display_axis is True:  # Display axis if key 'axis' not found (default display axis) or True
            axis_specs['y']['start'] = '0 0 0'
            axis_specs['y']['end'] = f'0 {self._max_height} 0'

            # Axis labels
            label_y_positions = np.linspace(
                start=self._max_height / Y_NUM_OF_TICKS,  # The lower label does not start in the ground
                stop=self._max_height,
                num=Y_NUM_OF_TICKS
            )
            label_values = np.linspace(  # Array of Y_NUM_OF_TICKS equally spaced values
                start=self._y_data.max() / Y_NUM_OF_TICKS,
                stop=self._y_data.max(),
                num=Y_NUM_OF_TICKS
            )

            x_coords = pd.Series(  # Series of Y_LABELS_X_DELTA values (repeated Y_NUM_OF_TICKS times)
                data=np.full(Y_NUM_OF_TICKS, Y_LABELS_X_DELTA),
                index=pd.RangeIndex(Y_NUM_OF_TICKS)
            )
            y_coords = pd.Series(label_y_positions, index=pd.RangeIndex(Y_NUM_OF_TICKS))
            z_coords = pd.Series(  # Series of 0 values (repeated Y_NUM_OF_TICKS times)
                data=np.full(Y_NUM_OF_TICKS, 0),
                index=pd.RangeIndex(Y_NUM_OF_TICKS)
            )
            label_pos_series = x_coords.astype(str).str.cat([y_coords.astype(str), z_coords.astype(str)], sep=' ')

            axis_specs['y']['labels_pos'] = label_pos_series.tolist()
            axis_specs['y']['labels_values'] = label_values.tolist()
            axis_specs['y']['labels_rotation'] = '0 0 0'

        # ---- Z-axis ----
        # Axis line
        display_axis = True
        try:
            display_axis = self._encoding['z']['axis'] if self._encoding.get('z') else False
        except KeyError or display_axis is True:  # Display axis if key 'axis' not found (default display axis) or True
            axis_specs['z']['start'] = '0 0 0'
            axis_specs['z']['end'] = f'0 0 {-DEFAULT_MAX_DEPTH}'

            # Axis labels
            categories = pd.Categorical(self._z_data)  # Transform categories into numbers (faster indexation)
            unique_label_values = categories.categories.tolist()  # Get only the unique values

            x_coords = pd.Series(  # Series of Z_LABELS_X_DELTA values
                data=np.full(len(categories.categories), Z_LABELS_X_DELTA),
                index=pd.RangeIndex(len(categories.categories))
            )
            y_coords = pd.Series(  # Series of LABELS_Y_DELTA values
                data=np.full(len(categories.categories), LABELS_Y_DELTA),
                index=pd.RangeIndex(len(categories.categories))
            )
            z_coords = self._set_z_coordinates()  # Z-axis coordinates for labels (aligned with bar centers)
            label_pos_series = x_coords.astype(str).str.cat([y_coords.astype(str), z_coords.astype(str)], sep=' ')

            axis_specs['z']['labels_pos'] = label_pos_series.tolist()
            axis_specs['z']['labels_values'] = unique_label_values
            axis_specs['z']['labels_rotation'] = '-90 0 0'

        return axis_specs


class GLTFModelCreator(ChartCreator):
    """GLTF model creator class."""

    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._url = chart_specs['data']['url']  # URL of the image model
        self._scale = chart_specs['mark'].get('scale', DEFAULT_GLTF_SCALE)

    def get_elements_specs(self) -> list[dict]:
        """Returns a list of dictionaries with the specifications for each element of the chart."""

        return [{'src': self._url, 'scale': self._scale}]

    def get_axis_specs(self) -> dict:
        """Returns a dictionary with the specifications for each axis of the chart."""

        axis_specs = copy.deepcopy(AXIS_DICT_TEMPLATE)
        return axis_specs  # GLTF models have no axis


class ImageCreator(ChartCreator):
    """Image creator class."""

    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._url = chart_specs['data']['url']  # URL of the image model
        self._height = chart_specs['mark'].get('height', DEFAULT_IMAGE_HEIGHT)
        self._width = chart_specs['mark'].get('width', DEFAULT_IMAGE_WIDTH)

    def get_elements_specs(self) -> list[dict]:
        """Returns a list of dictionaries with the specifications for each element of the chart."""

        return [{'src': self._url, 'width': self._width, 'height': self._height}]

    def get_axis_specs(self) -> dict:
        """Returns a dictionary with the specifications for each axis of the chart."""

        axis_specs = copy.deepcopy(AXIS_DICT_TEMPLATE)
        return axis_specs  # Images have no axis


class PointChartCreator(ChartCreator):
    """Point chart creator class."""

    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._raw_data = _get_raw_data(chart_specs)  # Raw data
        self._height = chart_specs.get('height', DEFAULT_MAX_HEIGHT)
        self._max_radius = chart_specs['mark'].get('max_radius', DEFAULT_POINT_RADIUS)
        self._color_data: Series | None = None
        self._size_data: Series | None = None
        self._x_data: Series | None = None
        self._y_data: Series | None = None
        self._z_data: Series | None = None

    def _set_points_colors(self) -> Series:
        """Returns a Series of the color for each point composing the scatter plot."""

        if self._color_data is None:
            raise Exception('Should never enter here.')

        categories = pd.Categorical(self._color_data)  # Transform categories to numbers (faster indexation)
        color_cycle = cycle(AVAILABLE_COLORS)  # Color cycle
        color_map_array = np.array(list(islice(color_cycle, len(categories.categories))))  # Array of colors (moduled)
        point_colors_array = color_map_array[categories.codes]  # Assign one color for each category
        points_colors = pd.Series(point_colors_array, index=self._color_data.index)
        return points_colors

    def _set_points_radius(self) -> Series:
        """Returns a Series of the radius for each point composing the bubble chart."""

        if self._size_data is None:
            raise Exception('Should never enter here.')

        max_value = self._size_data.max()
        points_radius_series = (self._size_data / max_value) * self._max_radius
        return points_radius_series

    def _set_x_coordinates(self, points_radius: Series) -> Series:
        """Returns a Series of the x coordinates for each point composing the point chart."""

        base_x = points_radius.iat[0]  # Take the radius of the first element so the chart starts in the base position
        if self._x_data is None:
            x_coordinates = Series(  # Series of base_x values
                data=np.full(len(self._raw_data), base_x),
                index=self._raw_data.index
            )
        else:
            x_coordinates = Series(
                data=base_x + (np.arange(len(self._x_data)) * DEFAULT_POINT_X_SEPARATION),
                index=self._raw_data.index
            )
        return x_coordinates

    def _set_y_coordinates(self, points_radius: Series) -> Series:
        """Returns a Series of the y coordinates for each point composing the point chart."""

        base_y = points_radius.max()  # Assert no points cross under the chart

        if self._y_data is None:
            y_coordinates = Series(  # Series of base_y values
                data=np.full(len(self._raw_data), base_y),
                index=self._raw_data.index
            )
        else:
            max_value = self._y_data.max()  # Proportional heights of the data
            y_coordinates = base_y + (self._y_data / max_value) * self._height  # Series of y-axis coordinates
        return y_coordinates

    def _set_z_coordinates(self) -> Series:
        """Returns a Series of the z coordinates for each point composing the point chart."""

        base_z = -DEFAULT_POINT_RADIUS

        if self._z_data is None:
            z_coordinates = Series(
                data=np.full(len(self._raw_data), base_z),
                index=self._raw_data.index
            )
        else:
            categories = pd.Categorical(self._z_data)  # Transform categories to numbers (faster indexation)
            z_coordinates_map = np.linspace(
                start=base_z,
                stop=-DEFAULT_MAX_DEPTH + DEFAULT_POINT_RADIUS,
                num=len(categories.categories)
            )
            z_coordinates = pd.Series(z_coordinates_map[categories.codes], index=self._z_data.index)
        return z_coordinates

    def get_elements_specs(self) -> list[dict]:
        """Returns a list of dictionaries with the specifications for each element of the chart."""

        if self._raw_data.empty:  # There is no data to display
            return []

        # X-axis
        radius = Series(  # Series of self._max_radius values
            data=np.full(len(self._raw_data), self._max_radius),
            index=self._raw_data.index
        )

        if self._encoding.get('x'):
            x_field = self._encoding['x']['field']
            try:
                self._x_data = self._raw_data[x_field]
            except KeyError:
                raise KeyError(f'Data has no field "{x_field}".')

            if self._encoding.get('size'):  # Bubbles plot (the size of the point depends on the value of the field)
                size_field = self._encoding['size']['field']
                try:
                    self._size_data = self._raw_data[size_field]
                except KeyError:
                    raise KeyError(f'Data has no field "{size_field}".')

                radius = self._set_points_radius()
            else:  # Scatter plot (same radius for all points)
                pass

        x_coordinates = self._set_x_coordinates(radius)

        # Y-axis
        if self._encoding.get('y'):
            y_field = self._encoding['y']['field']  # Field of the y-axis
            try:
                self._y_data = self._raw_data[y_field]
            except KeyError:
                raise KeyError(f'Data has no field "{y_field}".')

        y_coordinates = self._set_y_coordinates(radius)

        # Z-axis
        if self._encoding.get('z'):
            z_field = self._encoding['z']['field']  # Field of the z-axis
            try:
                self._z_data = self._raw_data[z_field]
            except KeyError:
                raise KeyError(f'Data has no field "{z_field}".')

        z_coordinates = self._set_z_coordinates()

        # Color
        if self._encoding.get('color'):  # Scatter plot (same color for each type of point)
            color_field = self._encoding['color']['field']
            try:
                self._color_data = self._raw_data[color_field]
            except KeyError:
                raise KeyError(f'Data has no field "{color_field}".')

            colors = self._set_points_colors()
        else:  # Bubbles plot (same color for all points)
            colors = Series(  # Series of DEFAULT_POINT_COLOR values
                data=np.full(len(self._raw_data), DEFAULT_POINT_COLOR),
                index=self._raw_data.index
            )

        # Id
        ids_series = []
        if self._x_data is not None:
            ids_series.append(self._x_data.astype(str))
        if self._y_data is not None:
            ids_series.append(self._y_data.astype(str))
        if self._z_data is not None:
            ids_series.append(self._z_data.astype(str))

        ids = ids_series[0].str.cat(others=ids_series[1:], sep=' : ', na_rep='?')  # Concatenate values of the series

        # Return values
        pos_series = x_coordinates.astype(str).str.cat([y_coordinates.astype(str), z_coordinates.astype(str)], sep=' ')
        temp_df = pd.DataFrame({
            'id': ids,
            'pos': pos_series,
            'radius': radius,
            'color': colors,
        })
        elements_specs = temp_df.to_dict(orient='records')  # Transform DataFrame into a list of dictionaries
        return elements_specs

    def get_axis_specs(self) -> dict:
        """Returns a dictionary with the specifications for each axis of the chart."""

        axis_specs = copy.deepcopy(AXIS_DICT_TEMPLATE)

        if self._raw_data.empty:  # There is no data to display
            return axis_specs

        # ---- X-axis ----
        # Axis line
        display_axis = True
        try:
            display_axis = self._encoding['x']['axis'] if self._encoding.get('x') else False
        except KeyError or display_axis is True:  # Display axis if key not found (default display axis) or True
            axis_specs['x']['start'] = '0 0 0'
            axis_specs['x']['end'] = f'{DEFAULT_POINT_X_SEPARATION * len(self._raw_data) + self._max_radius} 0 0'

            # Axis labels
            if self._encoding.get('size'):  # Bubbles plot (the size of the point depends on the value of the field)
                radius = self._set_points_radius()
            else:  # Scatter plot (same radius for all points)
                radius = Series(  # Series of self._max_radius values
                    data=np.full(len(self._raw_data), self._max_radius),
                    index=self._raw_data.index
                )

            x_coords = self._set_x_coordinates(radius)  # X-axis value for each point
            y_coords = Series(  # Series of LABELS_Y_DELTA values
                data=np.full(len(self._x_data), LABELS_Y_DELTA),
                index=self._x_data.index
            )
            z_coords = Series(  # Series of LABELS_Z_DELTA values
                data=np.full(len(self._x_data), LABELS_Z_DELTA),
                index=self._x_data.index
            )
            label_pos_series = x_coords.astype(str).str.cat([y_coords.astype(str), z_coords.astype(str)], sep=' ')

            axis_specs['x']['labels_pos'] = label_pos_series.tolist()
            axis_specs['x']['labels_values'] = self._x_data.tolist()
            axis_specs['x']['labels_rotation'] = '-90 0 -90'  # Rotation of the labels

        # ---- Y-axis ----
        # Axis line
        display_axis = True
        try:
            display_axis = self._encoding['y']['axis'] if self._encoding.get('y') else False
        except KeyError or display_axis is True:  # Display axis if key not found (default display axis) or True
            axis_specs['y']['start'] = '0 0 0'
            axis_specs['y']['end'] = f'0 {self._height} 0'

            # Axis labels
            label_y_positions = np.linspace(
                start=self._height / Y_NUM_OF_TICKS,
                stop=self._height,
                num=Y_NUM_OF_TICKS
            )
            label_values = np.linspace(  # Array of Y_NUM_OF_TICKS equally spaced values
                start=self._y_data.max() / Y_NUM_OF_TICKS,
                stop=self._y_data.max(),
                num=Y_NUM_OF_TICKS
            )

            x_coords = pd.Series(  # Series of Y_LABELS_X_DELTA values (repeated Y_NUM_OF_TICKS)
                data=np.full(Y_NUM_OF_TICKS, Y_LABELS_X_DELTA),
                index=pd.RangeIndex(Y_NUM_OF_TICKS)
            )
            y_coords = pd.Series(label_y_positions, index=pd.RangeIndex(Y_NUM_OF_TICKS))
            z_coords = pd.Series(  # Series of 0 values
                data=np.full(Y_NUM_OF_TICKS, 0),
                index=pd.RangeIndex(Y_NUM_OF_TICKS)
            )
            label_pos_series = x_coords.astype(str).str.cat([y_coords.astype(str), z_coords.astype(str)], sep=' ')

            axis_specs['y']['labels_pos'] = label_pos_series.tolist()
            axis_specs['y']['labels_values'] = label_values.tolist()
            axis_specs['y']['labels_rotation'] = '0 0 0'

        # ---- Z-axis ----
        # Axis line
        display_axis = True
        try:
            display_axis = self._encoding.get('z')['axis'] if self._encoding.get('z') else False
        except KeyError or display_axis is True:  # Display axis if key 'axis' not found (default display axis) or True
            axis_specs['z']['start'] = '0 0 0'
            axis_specs['z']['end'] = f'0 0 {-DEFAULT_MAX_DEPTH}'

            # Axis labels
            categories = pd.Categorical(self._z_data)  # Transform categories into numbers (faster indexation)
            label_z_positions = np.linspace(
                start=-(DEFAULT_POINT_RADIUS / 2),  # Start where the points have the center
                stop=-(DEFAULT_POINT_RADIUS / 2) - DEFAULT_MAX_DEPTH,  # End where the points have the center
                num=len(categories.categories)
            )
            unique_label_values = categories.categories.tolist()  # Get only the unique values

            x_coords = pd.Series(  # Series of Z_LABELS_X_DELTA values
                data=np.full(len(categories.categories), Z_LABELS_X_DELTA),
                index=pd.RangeIndex(len(categories.categories))
            )
            y_coords = pd.Series(  # Series of LABELS_Y_DELTA values
                data=np.full(len(categories.categories), LABELS_Y_DELTA),
                index=pd.RangeIndex(len(categories.categories))
            )
            z_coords = pd.Series(label_z_positions)
            label_pos_series = x_coords.astype(str).str.cat([y_coords.astype(str), z_coords.astype(str)], sep=' ')

            axis_specs['z']['labels_pos'] = label_pos_series.tolist()
            axis_specs['z']['labels_values'] = unique_label_values
            axis_specs['z']['labels_rotation'] = '-90 0 0'

        return axis_specs



"""AframeXR axis HTML creator"""


class AxisHTMLCreator:
    """Axis HTML creator class."""

    @staticmethod
    def create_axis_html(start: str | None, end: str | None) -> str:
        """
        Create a line for the axis and returns its HTML.

        Parameters
        ----------
        start : str | None
            The base position of each axis. If None, no axis is displayed.
        end : str | None
            The end position of the axis. If None, no axis is displayed.
        """

        if start and end:
            return f'<a-entity line="start: {start}; end: {end}; color: black"></a-entity>'
        return ''

    @staticmethod
    def create_label_html(pos: str, rotation: str, value: str) -> str:
        """
        Create a text with the value of the label in the correct position and returns its HTML.

        Parameters
        ----------
        pos : str
            The position of the label.
        rotation : str
            The rotation of the label (for better visualization).
        value : str
            The value of the label.
        """

        return f'<a-text position="{pos}" rotation="{rotation}" value="{value}" scale="3 3 3"></a-text>'



import aframexr
import math
import unittest

from bs4 import BeautifulSoup

from aframexr.api.filters import FilterTransform
from aframexr.utils.constants import DEFAULT_MAX_HEIGHT, DEFAULT_BAR_WIDTH
from tests.constants import *  # Constants used for testing


def _all_bars_have_same_width(bars_chart: aframexr.Chart) -> bool:
    """Verify that every bar has the same width."""

    width_for_every_bar = float(bars_chart.to_dict()['mark'].get('width', DEFAULT_BAR_WIDTH))

    soup = BeautifulSoup(bars_chart.to_html(), 'html.parser')
    bars = soup.find_all('a-box')
    for b in bars:
        bar_width = float(b['width'])  # Width of the bar
        if not math.isclose(bar_width, width_for_every_bar):
            return False
    return True

def _bars_bases_are_on_x_axis(bars_chart: aframexr.Chart) -> bool:
    """Verify that the bars are well-placed in the x-axis (the base of the bar is in the x-axis)."""

    soup = BeautifulSoup(bars_chart.to_html(), 'html.parser')
    bars = soup.find_all('a-box')
    for b in bars:
        bar_height = float(b['height'])  # Total height of the bar
        y_axis_midpoint = float(b['position'].split()[1])  # Y-axis coordinate
        if not math.isclose(bar_height, 2 * y_axis_midpoint):
            return False  # The height of the bar must be twice the value of its y-axis midpoint
    return True

def _bars_height_does_not_exceed_max_height(bars_chart: aframexr.Chart) -> bool:
    """Verify that every bar height does not exceed the maximum height."""

    max_height = float(bars_chart.to_dict().get('height', DEFAULT_MAX_HEIGHT))

    soup = BeautifulSoup(bars_chart.to_html(), 'html.parser')
    bars = soup.find_all('a-box')
    for b in bars:
        bar_height = float(b['height'])  # Total height of the bar
        if bar_height > max_height:
            return False
    return True


class TestMarkBarOK(unittest.TestCase):
    """Bars chart OK tests."""

    def test_simple(self):
        """Simple bars chart creation."""

        bars_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales')
        bars_chart.show()
        assert _all_bars_have_same_width(bars_chart)
        assert _bars_bases_are_on_x_axis(bars_chart)
        assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_data_format(self):
        """Bars chart changing data format creation."""

        for d in DATA_FORMATS:
            bars_chart = aframexr.Chart(d).mark_bar().encode(x='model', y='sales')
            bars_chart.show()
            assert _all_bars_have_same_width(bars_chart)
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_position(self):
        """Bars chart changing position creation."""

        for p in POSITIONS:
            bars_chart = aframexr.Chart(DATA, position=p).mark_bar().encode(x='model', y='sales')
            bars_chart.show()
            assert _all_bars_have_same_width(bars_chart)
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_position_format(self):
        """Bars chart changing position format creation."""

        for p in POSITION_FORMATS:
            bars_chart = aframexr.Chart(DATA, position=p).mark_bar().encode(x='model', y='sales')
            bars_chart.show()
            assert _all_bars_have_same_width(bars_chart)
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_rotation(self):
        """Bars chart changing rotation creation."""

        for r in ROTATIONS:
            bars_chart = aframexr.Chart(DATA, rotation=r).mark_bar().encode(x='model', y='sales')
            bars_chart.show()
            assert _all_bars_have_same_width(bars_chart)
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_rotation_format(self):
        """Bars chart changing rotation format creation."""

        for r in ROTATION_FORMATS:
            bars_chart = aframexr.Chart(DATA, rotation=r).mark_bar().encode(x='model', y='sales')
            bars_chart.show()
            assert _all_bars_have_same_width(bars_chart)
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_size(self):
        """Bars chart changing size creation."""

        for s in MARK_BAR_POINT_SIZES:
            bars_chart = aframexr.Chart(DATA).mark_bar(size=s).encode(x='model', y='sales')
            bars_chart.show()
            assert _all_bars_have_same_width(bars_chart)
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_height(self):
        """Bars chart changing height creation."""

        for h in MARK_BAR_POINT_HEIGHTS:
            bars_chart = aframexr.Chart(DATA).mark_bar(height=h).encode(x='model', y='sales')
            bars_chart.show()
            assert _all_bars_have_same_width(bars_chart)
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_encoding(self):
        """Bars chart changing encoding creation."""

        for e in MARK_BAR_ENCODINGS:
            bars_chart = aframexr.Chart(DATA).mark_bar().encode(**e)
            bars_chart.show()
            assert _all_bars_have_same_width(bars_chart)
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_filter(self):
        """Bars chart changing filter creation."""

        for eq in FILTER_EQUATIONS:
            for f in [eq, FilterTransform.from_string(eq)]:  # Filter using equation and using FilterTransform object
                bars_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').transform_filter(f)
                bars_chart.show()
                assert _all_bars_have_same_width(bars_chart)
                assert _bars_bases_are_on_x_axis(bars_chart)
                assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_position_rotation_size_height_encoding_filter(self):
        """Bars chart changing position, rotation size, height, encoding and filter creation."""

        for p, r, s, h, e, f in zip(POSITIONS, ROTATIONS, MARK_BAR_POINT_SIZES, MARK_BAR_POINT_HEIGHTS,
                                    MARK_BAR_ENCODINGS, FILTER_EQUATIONS):
            bars_chart = (aframexr.Chart(DATA, position=p, rotation=r).mark_bar(size=s, height=h).encode(**e)
                          .transform_filter(f))
            bars_chart.show()
            assert _all_bars_have_same_width(bars_chart)
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)


class TestMarkBarError(unittest.TestCase):
    """Bars chart error tests."""

    def test_position_error(self):
        """Bars chart position error."""

        for p in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_bar().encode(x='model', y='sales')
            assert str(error.exception) == f'The position: {p} is not correct. Must be "x y z".'

        for p in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_bar().encode(x='model', y='sales')
            assert str(error.exception) == 'The position values must be numeric.'

    def test_rotation_error(self):
        """Bars chart rotation error."""

        for r in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_bar().encode(x='model', y='sales')
            assert str(error.exception) == f'The rotation: {r} is not correct. Must be "x y z".'

        for r in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_bar().encode(x='model', y='sales')
            assert str(error.exception) == 'The rotation values must be numeric.'

    def test_size_error(self):
        """Bars chart size error."""

        for s in NOT_GREATER_THAN_0_MARK_BAR_POINT_SIZES_HEIGHTS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_bar(size=s).encode(x='model', y='sales')
            assert str(error.exception) == 'The size must be greater than 0.'

    def test_height_error(self):
        """Bars chart height error."""

        for h in NOT_GREATER_THAN_0_MARK_BAR_POINT_SIZES_HEIGHTS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_bar(height=h).encode(x='model', y='sales')
            assert str(error.exception) == 'The height must be greater than 0.'

    def test_encoding_error(self):
        """Bars chart encoding error."""

        for e in NON_EXISTING_MARK_BAR_POINT_ENCODINGS:
            with self.assertRaises(KeyError) as error:
                bars_chart = aframexr.Chart(DATA).mark_bar().encode(**e)
                bars_chart.show()
            assert 'Data has no field ' in str(error.exception)

        for e in NOT_VALID_MARK_BAR_POINT_ENCODINGS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_bar().encode(**e)
            assert str(error.exception) == 'At least 2 of (x, y, z) must be specified.'

    def test_filter_warning(self):
        """Bars chart filter warning."""

        for f in WARNING_FILTER_EQUATIONS:
            with self.assertWarns(UserWarning) as warning:
                filt_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').transform_filter(f)
                filt_chart.show()
            assert str(warning.warning) == f'Data does not contain values for the filter: {f}.'

    def test_filter_error(self):
        """Bars chart filter error."""

        for f in ERROR_FILTER_EQUATIONS:
            with self.assertRaises(SyntaxError) as error:
                filt_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').transform_filter(f)
                filt_chart.show()
            assert str(error.exception) in ['Incorrect syntax, must be datum.{field} == {value}',
                                            'Incorrect syntax, must be datum.{field} > {value}',
                                            'Incorrect syntax, must be datum.{field} < {value}']


import aframexr
import unittest

from tests.constants import *

DATA = aframexr.URLData('https://raw.githubusercontent.com/KhronosGroup/glTF-Sample-Assets/refs/heads/main/Models/'
                        'AntiqueCamera/glTF/AntiqueCamera.gltf')


class TestMarkGLTFOK(unittest.TestCase):
    """Mark GLTF OK tests."""

    def test_simple(self):
        """Simple GLTF creation."""

        aframexr.Chart(DATA).mark_gltf().show()

    def test_position(self):
        """GLTF changing position creation."""

        for p in POSITIONS:
            aframexr.Chart(DATA, position=p).mark_gltf().show()

    def test_position_format(self):
        """GLTF changing position format creation."""

        for p in POSITION_FORMATS:
            aframexr.Chart(DATA, position=p).mark_gltf().show()

    def test_rotation(self):
        """GLTF changing rotation creation."""

        for r in ROTATIONS:
            aframexr.Chart(DATA, rotation=r).mark_gltf().show()

    def test_rotation_format(self):
        """GLTF changing rotation format creation."""

        for r in ROTATION_FORMATS:
            aframexr.Chart(DATA, rotation=r).mark_gltf().show()

    def test_scale(self):
        """GLTF changing scale creation."""

        for s in MARK_GLTF_SCALES:
            aframexr.Chart(DATA).mark_gltf(scale=s).show()

    def test_position_rotation_scale(self):
        """GLTF changing position, rotation and scale creation."""

        for p, r, s in zip(POSITIONS, ROTATIONS, MARK_GLTF_SCALES):
            aframexr.Chart(DATA, position=p, rotation=r).mark_gltf(scale=s).show()


class TestMarkGLTFError(unittest.TestCase):
    """Mark GLTF error tests."""

    def test_position_error(self):
        """Mark GLTF position error."""

        for p in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_gltf()
            assert str(error.exception) == f'The position: {p} is not correct. Must be "x y z".'

        for p in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_gltf()
            assert str(error.exception) == 'The position values must be numeric.'

    def test_rotation_error(self):
        """Mark GLTF rotation error."""

        for r in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_gltf()
            assert str(error.exception) == f'The rotation: {r} is not correct. Must be "x y z".'

        for r in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_gltf()
            assert str(error.exception) == 'The rotation values must be numeric.'



"""Constant / default values utils file"""

# ----- CONSTANTS -----
AVAILABLE_AGGREGATES = ['max', 'mean', 'min', 'sum']
AVAILABLE_COLORS = ['red', 'green', 'blue', 'yellow', 'magenta', 'cyan']
AVAILABLE_ENCODING_TYPES = {'Q': 'quantitative', 'O': 'ordinal', 'N': 'nominal'}
Y_LABELS_X_DELTA = -1.75  # Variation in the x-axis between the labels and the axis (add to y-axis pos for label pos)
Z_LABELS_X_DELTA = -3  # Variation in the x-axis between the labels and the axis (add to z-axis pos for label pos)
LABELS_Y_DELTA = 0.01  # Variation in the y-axis between the labels and the axis (add to x and z axis pos for label pos)
LABELS_Z_DELTA = 1  # Variation in the y-axis between the labels and the axis (add to x-axis pos for label pos)
Y_NUM_OF_TICKS = 5  # Number of ticks in the y-axis

# ----- TEMPLATES -----
CHART_TEMPLATES = {
    'arc': ('<a-cylinder id="{id}" position="{pos}" radius="{radius}" theta-start="{theta_start}" '
            'theta-length="{theta_length}" material="color: {color}" data-raycastable></a-cylinder>'),
    'bar': ('<a-box id="{id}" position="{pos}" width="{width}" height="{height}" '
            'material="color: {color}" data-raycastable></a-box>'),
    'point': ('<a-sphere id="{id}" position="{pos}" radius="{radius}" material="color: {color}" data-raycastable>'
             '</a-sphere>')
}
IMAGES_TEMPLATES = {
    'gltf': '<a-gltf-model src="{src}" scale="{scale}"></a-gltf-model>',
    'image': '<a-image src="{src}" width="{width}" height="{height}"></a-image>'
}
ALL_TEMPLATES = {**CHART_TEMPLATES, **IMAGES_TEMPLATES}  # Grouped dictionary with all templates

# ----- DEFAULTS -----
# General
DEFAULT_CHART_POS = '0 0 0'  # Default position of the chart
DEFAULT_CHART_ROTATION = '0 0 0'  # Default chart rotation
DEFAULT_MAX_DEPTH = 10  # Default maximum depth of the chart
DEFAULT_MAX_HEIGHT = 10  # Default maximum height of the chart

# Bar chart
DEFAULT_BAR_DEPTH = 1  # Default bar depth
DEFAULT_BAR_HEIGHT_WHEN_NO_Y_AXIS = 1  # Default bar height (if not field for y-axis specified)
DEFAULT_BAR_WIDTH = 1  # Default bar width

# Pie chart
DEFAULT_PIE_RADIUS = 2  # Default radius of the pie chart
DEFAULT_PIE_ROTATION = '-90 0 0'  # Default pie chart rotation
DEFAULT_PIE_INNER_RADIUS = 0  # Default inner radius of the pie chart

# GLTF model
DEFAULT_GLTF_SCALE = '1 1 1'  # Default scale of the GLTF model

# Image
DEFAULT_IMAGE_HEIGHT = 1  # Default height of the image
DEFAULT_IMAGE_WIDTH = 1  # Default width of the image

# Point chart
DEFAULT_POINT_RADIUS = 1  # Default point radius
DEFAULT_POINT_X_SEPARATION = 2  # Default horizontal separation between points
DEFAULT_POINT_COLOR = "blue"  # Default point color



import aframexr
import unittest

from bs4 import BeautifulSoup

from aframexr.api.filters import FilterTransform
from aframexr.utils.constants import DEFAULT_POINT_RADIUS
from tests.constants import *  # Constants used for testing


def _every_radius_does_not_exceed_max_radius(point_chart: aframexr.Chart) -> bool:
    """Verify that every point radius does not exceed the maximum radius."""

    max_radius = point_chart.to_dict()['mark'].get('max_radius', DEFAULT_POINT_RADIUS)

    soup = BeautifulSoup(point_chart.to_html(), 'html.parser')
    points = soup.find_all('a-sphere')
    for p in points:
        point_radius = float(p['radius'])  # Radius of the sphere
        if point_radius > max_radius:
            return False
    return True


class TestMarkPointOK(unittest.TestCase):
    """Mark point OK tests."""

    def test_simple(self):
        """Simple mark point creation."""

        point_chart = aframexr.Chart(DATA).mark_point().encode(x='model', y='sales')
        point_chart.show()
        assert _every_radius_does_not_exceed_max_radius(point_chart)

    def test_data_format(self):
        """Mark point changing data format creation."""

        for d in DATA_FORMATS:
            point_chart = aframexr.Chart(d).mark_point().encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)

    def test_position(self):
        """Mark point changing position creation."""

        for p in POSITIONS:
            point_chart = aframexr.Chart(DATA, position=p).mark_point().encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)

    def test_position_format(self):
        """Mark point changing position format creation."""

        for p in POSITION_FORMATS:
            point_chart = aframexr.Chart(DATA, position=p).mark_point().encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)

    def test_rotation(self):
        """Mark point changing rotation creation."""

        for r in ROTATIONS:
            point_chart = aframexr.Chart(DATA, rotation=r).mark_point().encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)

    def test_rotation_format(self):
        """Mark point changing rotation format creation."""

        for r in ROTATION_FORMATS:
            point_chart = aframexr.Chart(DATA, rotation=r).mark_point().encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)

    def test_size(self):
        """Mark point changing size creation."""

        for s in MARK_BAR_POINT_SIZES:
            point_chart = aframexr.Chart(DATA).mark_point(size=s).encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)

    def test_height(self):
        """Mark point changing height creation."""

        for h in MARK_BAR_POINT_HEIGHTS:
            point_chart = aframexr.Chart(DATA).mark_point(height=h).encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)

    def test_encoding(self):
        """Mark point changing encoding creation."""

        for e in MARK_POINT_ENCODINGS:
            point_chart = aframexr.Chart(DATA).mark_point().encode(**e)
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)

    def test_filter(self):
        """Mark point changing filter creation."""

        for eq in FILTER_EQUATIONS:
            for f in [eq, FilterTransform.from_string(eq)]:  # Filter using equation and using FilterTransform object
                point_chart = aframexr.Chart(DATA).mark_point().encode(x='model', y='sales').transform_filter(f)
                point_chart.show()
                assert _every_radius_does_not_exceed_max_radius(point_chart)

    def test_position_rotation_size_height_encoding_filter(self):
        """Mark point changing position, rotation size, height, encoding and filter creation."""

        for p, r, s, h, e, f in zip(POSITIONS, ROTATIONS, MARK_BAR_POINT_SIZES, MARK_BAR_POINT_HEIGHTS,
                                 MARK_POINT_ENCODINGS, FILTER_EQUATIONS):
            point_chart = (aframexr.Chart(DATA, position=p, rotation=r).mark_point(size=s, height=h).encode(**e)
                           .transform_filter(f))
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)

    def test_concatenating_charts(self):
        """Mark point concatenating charts creation."""

        for p, r, s, h, e, f in zip(POSITIONS, ROTATIONS, MARK_BAR_POINT_SIZES, MARK_BAR_POINT_HEIGHTS,
                                    MARK_POINT_ENCODINGS, FILTER_EQUATIONS):
            point_chart = (aframexr.Chart(DATA, position=p, rotation=r).mark_point(size=s, height=h).encode(**e)
                           .transform_filter(f))
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)


class TestMarkPointError(unittest.TestCase):
    """Mark point error tests."""

    def test_position_error(self):
        """Mark point position error."""

        for p in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_point().encode(x='model', y='sales')
            assert str(error.exception) == f'The position: {p} is not correct. Must be "x y z".'

        for p in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_point().encode(x='model', y='sales')
            assert str(error.exception) == 'The position values must be numeric.'

    def test_rotation_error(self):
        """Mark point rotation error."""

        for r in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_point().encode(x='model', y='sales')
            assert str(error.exception) == f'The rotation: {r} is not correct. Must be "x y z".'

        for r in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_point().encode(x='model', y='sales')
            assert str(error.exception) == 'The rotation values must be numeric.'

    def test_size_error(self):
        """Mark point size error."""

        for s in NOT_GREATER_THAN_0_MARK_BAR_POINT_SIZES_HEIGHTS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_point(size=s).encode(x='model', y='sales')
            assert str(error.exception) == 'The size must be greater than 0.'

    def test_height_error(self):
        """Mark point height error."""

        for h in NOT_GREATER_THAN_0_MARK_BAR_POINT_SIZES_HEIGHTS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_point(height=h).encode(x='model', y='sales')
            assert str(error.exception) == 'The height must be greater than 0.'

    def test_encoding_error(self):
        """Mark point encoding error."""

        for e in NON_EXISTING_MARK_BAR_POINT_ENCODINGS:
            with self.assertRaises(KeyError) as error:
                point_chart = aframexr.Chart(DATA).mark_point().encode(**e)
                point_chart.show()
            assert 'Data has no field ' in str(error.exception)

        for e in NOT_VALID_MARK_BAR_POINT_ENCODINGS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_point().encode(**e)
            assert str(error.exception) == 'At least 2 of (x, y, z) must be specified.'

    def test_filter_warning(self):
        """Mark point filter warning."""

        for f in WARNING_FILTER_EQUATIONS:
            with self.assertWarns(UserWarning) as warning:
                filt_chart = aframexr.Chart(DATA).mark_point().encode(x='model', y='sales').transform_filter(f)
                filt_chart.show()
            assert str(warning.warning) == f'Data does not contain values for the filter: {f}.'

    def test_filter_error(self):
        """Mark point filter error."""

        for f in ERROR_FILTER_EQUATIONS:
            with self.assertRaises(SyntaxError) as error:
                filt_chart = aframexr.Chart(DATA).mark_point().encode(x='model', y='sales').transform_filter(f)
                filt_chart.show()
            assert str(error.exception) in ['Incorrect syntax, must be datum.{field} == {value}',
                                            'Incorrect syntax, must be datum.{field} > {value}',
                                            'Incorrect syntax, must be datum.{field} < {value}']



"""AframeXR validators"""

from typing import Union

from aframexr.utils.constants import ALL_TEMPLATES, AVAILABLE_AGGREGATES, AVAILABLE_ENCODING_TYPES


class AframeXRValidator:
    """AframeXR validator class."""

    @staticmethod
    def validate_aggregate_operation(aggregate: str) -> None:
        """Raises ValueError if the aggregate operation is invalid."""

        if aggregate not in AVAILABLE_AGGREGATES:
            raise ValueError(f'Invalid aggregate operation: {aggregate}.')

    @staticmethod
    def validate_chart_type(chart_type: str) -> None:
        """Raises ValueError if the chart type is invalid."""

        if chart_type not in ALL_TEMPLATES:
            raise ValueError(f'Invalid chart type: {chart_type}.')

    @staticmethod
    def validate_encoding_type(encoding_type: str) -> None:
        """Raises ValueError if encoding type is not valid."""

        if not isinstance(encoding_type, str):
            raise TypeError(f'Expected str, got {type(encoding_type).__name__} instead.')

        if encoding_type not in AVAILABLE_ENCODING_TYPES:
            raise ValueError(f'Invalid encoding type: {encoding_type}.')

    @staticmethod
    def validate_there_are_no_charts_defined(chart_specs: dict) -> None:
        """Raises ValueError if there is a chart defined in the specifications."""

        if chart_specs.get('mark') is not None:  # There is one chart defined already
            raise ValueError('Trying to mark a chart already defined in the specifications.')

    @staticmethod
    def validate_type(param, types: Union[type]) -> None:
        """Raises TypeError if type(param) is not in types."""

        if not isinstance(param, types):
            raise TypeError(f'Expected {types.__name__}, got {type(param).__name__} instead.')


"""AframeXR encoding classes"""

from abc import ABC, abstractmethod
from typing import Union

from aframexr.utils.constants import AVAILABLE_ENCODING_TYPES
from aframexr.utils.validators import AframeXRValidator


class Encoding(ABC):
    """Encoding base class."""

    # Export
    @abstractmethod
    def to_dict(self):
        pass  # Must be implemented by child classes

    # Utils
    @staticmethod
    def split_field_and_encoding(param: str) -> tuple[str, str | None]:
        """
        Splits and returns the field and the encoding data type of the parameter.

        Raises
        ------
        TypeError
            If the encoding type is incorrect.

        Notes
        -----
        Supposing that param is a string, as it has been called from encode() method.
        """

        param_parts = param.split(':')  # Split parameter in field:encoding_type
        if len(param_parts) == 1:  # No encoding data type is specified
            return param, None
        if len(param_parts) == 2:
            field = param_parts[0]
            encoding_type = param_parts[1].upper()  # Convert to upper case (to accept lower case also)
            AframeXRValidator.validate_encoding_type(encoding_type)
            return field, AVAILABLE_ENCODING_TYPES[encoding_type]
        else:
            raise ValueError(f'Invalid encoding type: {param}.')


class X(Encoding):
    """
    X-axis encoding class.

    Parameters
    ----------
    field : str
        The data field of the axis.
    aggregate : bool | None (optional)
        Type of transformation in the field data.
    axis : bool | None (optional)
        If the axis line is visible or not. Default is True (visible).
    encoding_type : str | None (optional)
        The encoding type.
    group_by : str | None (optional)
        The grouping key to use for the encoding.
    """

    def __init__(self, field: str, aggregate: str | None = None, axis: bool | None = True,
                 encoding_type: str | None = None, group_by: str | None = None):
        AframeXRValidator.validate_type(field, str)
        self.field = field

        AframeXRValidator.validate_type(aggregate, Union[str | None])
        AframeXRValidator.validate_aggregate_operation(aggregate)
        self.aggregate = aggregate

        AframeXRValidator.validate_type(axis, Union[bool | None])
        self.axis = axis

        AframeXRValidator.validate_type(encoding_type, Union[str | None])
        AframeXRValidator.validate_encoding_type(encoding_type)
        self.encoding_type = encoding_type

        AframeXRValidator.validate_type(group_by, str)
        self.group_by = group_by

    # Export
    def to_dict(self):
        """Returns the dictionary specifications expression."""

        spec_dict = {'x': {}}
        if self.field:
            spec_dict['x']['field'] = self.field
        if self.aggregate:
            spec_dict['x']['aggregate'] = self.aggregate
        if not self.axis:  # Add if it is not True (as True is the default)
            spec_dict['x']['axis'] = self.axis
        if self.encoding_type:
            spec_dict['x']['encoding_type'] = self.encoding_type
        if self.group_by:
            spec_dict['x']['group_by'] = self.group_by
        return spec_dict


class Y(Encoding):
    """
    Y-axis encoding class.

    Parameters
    ----------
    field : str
        The data field of the axis.
    aggregate : bool | None (optional)
        Type of transformation in the field data.
    axis : bool | None (optional)
        If the axis line is visible or not. Default is True (visible).
    encoding_type : str | None (optional)
        The encoding type.
    group_by : str | None (optional)
        The grouping key to use for the encoding.
    """

    def __init__(self, field: str, aggregate: str | None = None, axis: bool | None = True,
                 encoding_type: str | None = None, group_by: str | None = None):
        AframeXRValidator.validate_type(field, str)
        self.field = field

        AframeXRValidator.validate_type(aggregate, Union[str | None])
        AframeXRValidator.validate_aggregate_operation(aggregate)
        self.aggregate = aggregate

        AframeXRValidator.validate_type(axis, Union[bool | None])
        self.axis = axis

        AframeXRValidator.validate_type(encoding_type, Union[str | None])
        AframeXRValidator.validate_encoding_type(encoding_type)
        self.encoding_type = encoding_type

        AframeXRValidator.validate_type(group_by, str)
        self.group_by = group_by

    # Export
    def to_dict(self):
        """Returns the dictionary specifications expression."""

        spec_dict = {'y': {}}
        if self.field:
            spec_dict['y']['field'] = self.field
        if self.aggregate:
            spec_dict['y']['aggregate'] = self.aggregate
        if not self.axis:  # Add if it is not True (as True is the default)
            spec_dict['y']['axis'] = self.axis
        if self.encoding_type:
            spec_dict['y']['encoding_type'] = self.encoding_type
        if self.group_by:
            spec_dict['y']['group_by'] = self.group_by
        return spec_dict


class Z(Encoding):
    """
    Z-axis encoding class.

    Parameters
    ----------
    field : str
        The data field of the axis.
    aggregate : bool | None (optional)
        Type of transformation in the field data.
    axis : bool | None (optional)
        If the axis line is visible or not. Default is True (visible).
    encoding_type : str | None (optional)
        The encoding type.
    group_by : str | None (optional)
        The grouping key to use for the encoding.
    """

    def __init__(self, field: str, aggregate: str | None = None, axis: bool | None = True,
                 encoding_type: str | None = None, group_by: str | None = None):
        AframeXRValidator.validate_type(field, str)
        self.field = field

        AframeXRValidator.validate_type(aggregate, Union[str | None])
        AframeXRValidator.validate_aggregate_operation(aggregate)
        self.aggregate = aggregate

        AframeXRValidator.validate_type(axis, Union[bool | None])
        self.axis = axis

        AframeXRValidator.validate_type(encoding_type, Union[str | None])
        AframeXRValidator.validate_encoding_type(encoding_type)
        self.encoding_type = encoding_type

        AframeXRValidator.validate_type(group_by, str)
        self.group_by = group_by

    # Export
    def to_dict(self):
        """Returns the dictionary specifications expression."""

        spec_dict = {'z': {}}
        if self.field:
            spec_dict['z']['field'] = self.field
        if self.aggregate:
            spec_dict['z']['aggregate'] = self.aggregate
        if not self.axis:  # Add if it is not True (as True is the default)
            spec_dict['z']['axis'] = self.axis
        if self.encoding_type:
            spec_dict['z']['encoding_type'] = self.encoding_type
        return spec_dict



"""AframeXR components"""

import copy
import json
import marimo

from pandas import DataFrame
from typing import Literal, Union

from aframexr.api.aggregate import AggregatedFieldDef
from aframexr.api.data import Data, URLData
from aframexr.api.encoding import Encoding, X, Y, Z
from aframexr.api.filters import FilterTransform
from aframexr.utils.constants import *
from aframexr.utils.scene_creator import SceneCreator
from aframexr.utils.validators import AframeXRValidator


class TopLevelMixin:
    """Top level chart class."""

    def __init__(self, specs: dict = None):
        if specs is None:  # For calls of Chart.__init__(), calling super().__init__()
            self._specifications = {}  # Specifications of the scene, in JSON format
        else:  # For calls of __add__(), to create a new object and do not modify the rest
            self._specifications = specs

    # Concatenating charts
    def __add__(self, other):
        """
        Concatenation of charts (place charts in the same scene).
        Creates and returns a new scene with the charts. The original charts are not modified.
        """

        if not isinstance(other, TopLevelMixin):
            raise TypeError(f"Cannot add {type(other).__name__} to {type(self).__name__}.")

        # Create the new concatenation specifications
        concat_specs = {'concat': []}

        # Look if there is concatenation of concatenated charts (to join all in one concat field)
        if self._specifications.get('concat'):
            for chart in self._specifications['concat']:
                concat_specs['concat'].append(chart)
        else:
            concat_specs['concat'].append(self._specifications)

        if other._specifications.get('concat'):
            for chart in other._specifications['concat']:
                concat_specs['concat'].append(chart)
        else:
            concat_specs['concat'].append(other._specifications)

        concatenated_chart = TopLevelMixin(concat_specs)  # Create a new 'scene' to preserve the original charts
        return concatenated_chart

    # Copy of the chart
    def copy(self):
        """Returns a deep copy of the chart."""

        return copy.deepcopy(self)

    # Importing charts
    @staticmethod
    def from_dict(specs: dict) -> 'TopLevelMixin':
        """
        Import the chart from the JSON dict specifications.

        Parameters
        ----------
        specs : dict
            JSON specifications of the chart.

        Raises
        ------
        TypeError
            If specs is not a dictionary.
        """

        AframeXRValidator.validate_type(specs, dict)
        return TopLevelMixin(specs)

    @staticmethod
    def from_json(specs: str) -> 'TopLevelMixin':
        """
        Create the chart from the JSON string specifications.

        Parameters
        ----------
        specs : str
            JSON specifications of the chart.

        Raises
        ------
        TypeError
            If specs is not a string.
        """

        AframeXRValidator.validate_type(specs, str)
        return TopLevelMixin(json.loads(specs))

    # Exporting charts
    def save(self, fp: str, fileFormat: Literal['json', 'html'] = None):
        """
        Saves the chart into a file, supported formats are JSON and HTML.

        Parameters
        ----------
        fp : str
            File path.
        fileFormat : str (optional)
            Format of the file could be ['html', 'json'].
            If no format is specified, the chart will be saved depending on the file extension.

        Raises
        ------
        ValueError
            If fileFormat is invalid.
        """

        AframeXRValidator.validate_type(fp, str)
        if fileFormat == 'html' or fp.endswith('.html'):
            with open(fp, 'w') as file:
                file.write(self.to_html())
        elif fileFormat == 'json' or fp.endswith('.json'):
            with open(fp, 'w') as file:
                json.dump(self._specifications, file, indent=4)
        else:
            raise ValueError('Invalid file format.')

    # Showing the scene
    def show(self):
        """Show the scene in the Marimo notebook."""

        html_scene = SceneCreator.create_scene(self._specifications)
        return marimo.iframe(html_scene)

    # Chart formats
    def to_dict(self) -> dict:
        """Returns the scene specifications as a dictionary."""

        return self._specifications

    def to_html(self) -> str:
        """Returns the HTML representation of the scene."""

        return SceneCreator.create_scene(self._specifications)

    def to_json(self) -> str:
        """Returns the JSON string of the scene."""

        return json.dumps(self._specifications)


class Chart(TopLevelMixin):
    """
    Simple chart class.

    Parameters
    ----------
    data : Data | URLData
        Data or URLData object of the data.

    position : str (optional)
        Position of the chart. The format is: 'x y z'. Refers to the position for the origin of coordinate system.
    rotation : str (optional)
        Rotation of the chart in degrees. The format is: 'x y z'. The rotation axis is the coordinate system.

    Raises
    ------
    TypeError
        If data is not a Data or URLData object.
    ValueError
        If position or rotation are invalid.
    """

    def __init__(self, data: Data | URLData | DataFrame, position: str = DEFAULT_CHART_POS,
                 rotation: str = DEFAULT_CHART_ROTATION):
        super().__init__()

        # Data
        if isinstance(data, Data):
            self._specifications.update({'data': {'values': data.values}})
        elif isinstance(data, URLData):
            self._specifications.update({'data': {'url': data.url}})
        elif isinstance(data, DataFrame):
            self._specifications.update({'data': {'values': data.to_dict(orient='records')}})
        else:
            raise TypeError(f'Expected Data | URLData | DataFrame, got {type(data).__name__} instead.')

        # Position
        pos_axes = position.strip().split()
        if len(pos_axes) != 3:
            raise ValueError(f'The position: {position} is not correct. Must be "x y z".')
        for axis in pos_axes:
            try:
                float(axis)
            except ValueError:
                raise ValueError('The position values must be numeric.')
        self._specifications.update({'position': f'{pos_axes[0]} {pos_axes[1]} {pos_axes[2]}'})

        # Rotation
        rot_axes = rotation.strip().split()
        if len(rot_axes) != 3:
            raise ValueError(f'The rotation: {rotation} is not correct. Must be "x y z".')
        for axis in rot_axes:
            try:
                float(axis)
            except ValueError:
                raise ValueError('The rotation values must be numeric.')
        self._specifications.update({'rotation': f'{rot_axes[0]} {rot_axes[1]} {rot_axes[2]}'})

    # Types of charts
    def _there_is_one_chart_defined(self):
        """Returns True if there is a chart defined in the specifications, False otherwise."""

        return self._specifications.get('mark') is not None

    def mark_arc(self, radius: float = DEFAULT_PIE_RADIUS):
        """
        Pie chart and doughnut chart.

        Parameters
        ----------
        radius : float (optional)
            Outer radius of the pie chart. If not specified, using default. Must be greater than 0.
        """

        AframeXRValidator.validate_type(radius, Union[float | int])
        AframeXRValidator.validate_there_are_no_charts_defined(self._specifications)

        self._specifications.update({'mark': {'type': 'arc'}})
        if radius > 0:
            self._specifications['mark'].update({'radius': radius})
        else:
            raise ValueError('The radius must be greater than 0.')
        return self

    def mark_bar(self, size: float = DEFAULT_BAR_WIDTH, height: float = DEFAULT_MAX_HEIGHT):
        """
        Bars chart.

        Parameters
        ----------
        size : float (optional)
            Width of the bars. If not specified, using default. Must be greater than 0.
        height : float (optional)
            Maximum height of the chart (the highest bar). If not specified, using default. Must be greater than 0.
        """

        AframeXRValidator.validate_type(size, Union[float | int])
        AframeXRValidator.validate_type(height, Union[float | int])
        AframeXRValidator.validate_there_are_no_charts_defined(self._specifications)

        self._specifications.update({'mark': {'type': 'bar'}})
        if size > 0:
            self._specifications['mark'].update({'width': size})
        else:
            raise ValueError('The size must be greater than 0.')
        if height > 0:
            self._specifications.update({'height': height})
        else:
            raise ValueError('The height must be greater than 0.')
        return self

    def mark_gltf(self, scale: str = DEFAULT_GLTF_SCALE):
        """
        GLTF model.

        Parameters
        ----------
        scale : str (optional)
            Scale of the GLTF model (from its original scale).

            **Format: 'x y z'** (values can be negative, works like a mirror).

            If an axis value is not specified, that value will be 1 (for example, '2 2' is the same as '2 2 1').

            If more than 3 axes are specified, then the first 3 axes will be used.
        """

        AframeXRValidator.validate_type(scale, str)
        AframeXRValidator.validate_there_are_no_charts_defined(self._specifications)

        self._specifications.update({'mark': {'type': 'gltf'}})
        self._specifications['mark'].update({'scale': scale})
        return self

    def mark_image(self, width: float = DEFAULT_IMAGE_WIDTH, height: float = DEFAULT_IMAGE_HEIGHT):
        """
        Image.

        Parameters
        ----------
        width : float (optional)
            Width of the image. If not specified, using default. Must be greater than 0.
        height : float (optional)
            Height of the image. If not specified, using default. Must be greater than 0.

        Raises
        ------
        ValueError
            If width or height are not greater than 0.
        """

        AframeXRValidator.validate_type(width, Union[float | int])
        AframeXRValidator.validate_type(height, Union[float | int])
        AframeXRValidator.validate_there_are_no_charts_defined(self._specifications)

        self._specifications.update({'mark': {'type': 'image'}})
        if width > 0:
            self._specifications['mark'].update({'width': width})
        else:
            raise ValueError('The width must be greater than 0.')
        if height > 0:
            self._specifications['mark'].update({'height': height})
        else:
            raise ValueError('The height must be greater than 0.')
        return self

    def mark_point(self, size: float = DEFAULT_POINT_RADIUS, height: float = DEFAULT_MAX_HEIGHT):
        """
        Scatter plot and bubble chart.

        Parameters
        ----------
        size : float (optional)
            Maximum radius of the point. If not specified, using default. Must be greater than 0.
        height : float (optional)
            Maximum height of the chart. If not specified, using default. Must be greater than 0.

        Raises
        ------
        ValueError
            If size or height are not greater than 0.
        """

        AframeXRValidator.validate_type(size, Union[float | int])
        AframeXRValidator.validate_type(height, Union[float | int])
        AframeXRValidator.validate_there_are_no_charts_defined(self._specifications)

        self._specifications.update({'mark': {'type': 'point'}})
        if size > 0:
            self._specifications['mark'].update({'max_radius': size})
        else:
            raise ValueError('The size must be greater than 0.')
        if height > 0:
            self._specifications.update({'height': height})
        else:
            raise ValueError('The height must be greater than 0.')
        return self

    # Parameters of the chart
    def encode(self, color: str = '', size: str = '', theta: str = '', x: str | X = '', y: str | Y = '',
               z: str | Z = ''):
        """
        Add properties to the chart.

        Parameters
        ----------
        color : str (optional)
            Field of the data that will determine the color of the sphere in the scatter plot.
        size : str (optional)
            Field of the data that will determine the size of the sphere in the bubble chart (must be quantitative).
        theta : str (optional)
            Field of the data that will determine the arcs of the pie and doughnut chart (must be quantitative).
        x : str | X (optional)
            Field of the data that will determine the x-axis of the chart.
        y : str | Y (optional)
            Field of the data what will determine the y-axis of the chart (must be quantitative).
        z : str | Z (optional)
            Field of the data what will determine the z-axis of the chart.

        Raises
        ------
        TypeError
            If the encoding type is incorrect.
        ValueError
            If the encoding values are incorrect.
        """

        filled_params = {}  # Dictionary that will store the parameters that have been filled

        # Verify the type of the arguments and store the filled parameters
        if color:
            AframeXRValidator.validate_type(color, str)
            filled_params.update({'color': color})
        if size:
            AframeXRValidator.validate_type(size, str)
            filled_params.update({'size': size})
        if theta:
            AframeXRValidator.validate_type(theta, str)
            filled_params.update({'theta': theta})
        if x:
            AframeXRValidator.validate_type(x, Union[str | X])
            filled_params.update({'x': x})
        if y:
            AframeXRValidator.validate_type(y, Union[str | Y])
            filled_params.update({'y': y})
        if z:
            AframeXRValidator.validate_type(z, Union[str | Z])
            filled_params.update({'z': z})

        # Verify the argument combinations
        if self._specifications['mark']['type'] in ['bar', 'point'] and sum([x != '', y != '', z != '']) < 2:
            raise ValueError('At least 2 of (x, y, z) must be specified.')
        if self._specifications['mark']['type'] == 'arc' and (not theta or not color):
            if not theta: raise ValueError('Parameter theta must be specified in arc chart.')
            if not color: raise ValueError('Parameter color must be specified in arc chart.')

        # Do the encoding
        self._specifications.update({'encoding': {}})
        for param_key in filled_params:
            param_value = filled_params[param_key]
            if isinstance(param_value, Encoding):
                self._specifications['encoding'].update(param_value.to_dict())
            else:
                formula, encoding_type = Encoding.split_field_and_encoding(param_value)
                field, aggregate_op = AggregatedFieldDef.split_operator_field(formula)

                self._specifications['encoding'].update({param_key: {'field': field}})
                if aggregate_op:
                    self._specifications['encoding'][param_key].update({'aggregate': aggregate_op})
                if encoding_type:
                    self._specifications['encoding'][param_key].update({'type': encoding_type})
        return self

    # Modifying data
    def transform_aggregate(self, groupby: list | None = None, **kwargs):
        """
        Aggregates the data with the specified aggregate function, grouped by the specified groupby.

        Parameters
        ----------
        groupby : list | None
            Data fields that will be grouped, optional. If not set, the defined fields in encode() method will be taken.
        kwargs : dict
            Format is: <new_field>=<aggregate_op>(<data_field>).
        """

        AframeXRValidator.validate_type(groupby, Union[list | None])

        # Create a copy of the chart (in case of assignation, to preserve the main chart)
        aggreg_chart = self.copy()

        if len(kwargs) == 0:  # At least one aggregation needs to be defined in kwargs
            raise TypeError('transform_aggregate() missing required aggregate fields.')

        aggregates_to_dict = []
        for as_field, aggregate_formula in kwargs.items():
            field, aggregate_op = AggregatedFieldDef.split_operator_field(str(aggregate_formula))
            aggregate_object = AggregatedFieldDef(aggregate_op, field, as_field)
            aggregates_to_dict.append(aggregate_object.to_dict())

        aggregate_specs = {'aggregate': aggregates_to_dict}
        if groupby:
            aggregate_specs['groupby'] = groupby

        if not aggreg_chart._specifications.get('transform'):  # First time filtering the chart (create field)
            aggreg_chart._specifications.update({'transform': [aggregate_specs]})
        else:  # Not the first filter of the chart (add to aggregates)
            aggreg_chart._specifications['transform'].append(aggregate_specs)
        return aggreg_chart

    def transform_filter(self, equation_filter: str | FilterTransform):
        """
        Filters the chart with the given transformation.

        Parameters
        ----------
        equation_filter : str | FilterTransform
            The equation string of the filter transformation, or a Filter object (see Examples).

        Raises
        ------
        TypeError
            If equation is not a string or a Filter object.

        Notes
        -----
        Can be concatenated with the rest of functions of the Chart, without needing an asignation. It can also be
        concatenated several times (the result will be an addition of the filters, in order of assignation).

        If the result of the filters is empty data, a warning is raised.

        Examples
        --------
        *Using transform_filter() giving the equation string:*

        >>> import aframexr
        >>> data = aframexr.URLData('./data.json')
        >>> chart = aframexr.Chart(data).mark_bar().encode(x='model', y='sales')
        >>> filtered_chart = chart.transform_filter('datum.motor == diesel')
        >>> #filtered_chart.show()

        *Using transform_filter() giving a Filter object*

        >>> import aframexr
        >>> data = aframexr.URLData('./data.json')
        >>> chart = aframexr.Chart(data).mark_bar().encode(x='model', y='sales')
        >>> filter_object = aframexr.FieldEqualPredicate(field='motor', equal='diesel')
        >>> filtered_chart = chart.transform_filter(filter_object)
        >>> #filtered_chart.show()
        """

        # Validate the type of equation_filter and get a filter object from the equation_filter
        if isinstance(equation_filter, str):
            filter_transform = FilterTransform.from_string(equation_filter)
        elif isinstance(equation_filter, FilterTransform):
            filter_transform = equation_filter
        else:
            raise TypeError(f'Expected str | FilterTransform, got {type(equation_filter).__name__} instead.')

        # Create a copy of the chart (in case of assignation, to preserve the main chart)
        filt_chart = self.copy()

        # Add the information of the filter object to the specifications
        if not filt_chart._specifications.get('transform'):  # First time filtering the chart
            filt_chart._specifications.update({'transform': [filter_transform.equation_to_dict()]})  # Create field
        else:  # Not the first filter of the chart
            filt_chart._specifications['transform'].append(filter_transform.equation_to_dict())  # Add filter to field
        return filt_chart  # Returns the copy of the chart



"""AframeXR scene creator"""

from aframexr.utils.entities_html_creator import ChartsHTMLCreator

HTML_SCENE_TEMPLATE = """<!DOCTYPE html>
<head>
    <script src="https://aframe.io/releases/1.7.1/aframe.min.js"></script>
    <script src="https://unpkg.com/aframe-environment-component@1.5.0/dist/aframe-environment-component.min.js"></script>
    <script src="https://davidlab20.github.io/TFG/scripts/main.js"></script>
</head>
<body>
    <a-scene cursor="rayOrigin: mouse" raycaster="objects: [data-raycastable]">
    
        <!-- Camera -->
        <a-camera position="0 4 0" active="true"></a-camera>
    
        <!-- Environment -->
        <a-entity environment="preset: default"></a-entity>
        
        <!-- Elements -->
        {elements}
        
        <!-- Variable label -->
        <a-plane id="plainLabel" position="" height="1.75" width="3.75" visible="false"></a-plane>
        <a-text id="label" position="" value="" scale="2 2 2" visible="false"></a-text>
    </a-scene>
</body>
"""


class SceneCreator:

    @staticmethod
    def create_scene(specs: dict):
        """
        Creates the HTML scene from the JSON specifications.

        Parameters
        ----------
        specs : dict
            Specifications of the elements composing the scene.

        Raises
        ------
        TypeError
            If specs is not a dictionary.

        Notes
        -----
        Suppose that specs is a dictionary for posterior method calls of ChartsHTMLCreator.
        """

        if not isinstance(specs, dict):
            raise TypeError(f'Expected specs to be a dict, got {type(specs).__name__}')
        elements_html = ChartsHTMLCreator.create_charts_html(specs)
        return HTML_SCENE_TEMPLATE.format(elements=elements_html)



import aframexr
import unittest

from tests.constants import *

DATA = aframexr.URLData('https://davidlab20.github.io/TFG/imgs/logo.png')


class TestMarkImageOK(unittest.TestCase):
    """Mark image OK tests."""

    def test_simple(self):
        """Simple image creation."""

        aframexr.Chart(DATA).mark_image().show()

    def test_position(self):
        """Image changing position creation."""

        for p in POSITIONS:
            aframexr.Chart(DATA, position=p).mark_image().show()

    def test_position_format(self):
        """Image changing position format creation."""

        for p in POSITION_FORMATS:
            aframexr.Chart(DATA, position=p).mark_image().show()

    def test_rotation(self):
        """Image changing rotation creation."""

        for r in ROTATIONS:
            aframexr.Chart(DATA, rotation=r).mark_image().show()

    def test_rotation_format(self):
        """Image changing rotation format creation."""

        for r in ROTATION_FORMATS:
            aframexr.Chart(DATA, rotation=r).mark_image().show()

    def test_width(self):
        """Image changing width creation."""

        for w in MARK_IMAGE_WIDTHS_HEIGHTS:
            aframexr.Chart(DATA).mark_image(width=w).show()

    def test_height(self):
        """Image changing height creation."""

        for h in MARK_IMAGE_WIDTHS_HEIGHTS:
            aframexr.Chart(DATA).mark_image(height=h).show()

    def test_position_rotation_width_height(self):
        """Image changing position, rotation, width and height creation."""

        for p, r, w, h in zip(POSITIONS, ROTATIONS, MARK_IMAGE_WIDTHS_HEIGHTS, MARK_IMAGE_WIDTHS_HEIGHTS):
            aframexr.Chart(DATA, position=p, rotation=r).mark_image(width=w, height=h).show()


class TestMarkImageError(unittest.TestCase):
    """Mark image error tests."""

    def test_position_error(self):
        """Mark image position error."""

        for p in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_image()
            assert str(error.exception) == f'The position: {p} is not correct. Must be "x y z".'

        for p in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_image()
            assert str(error.exception) == 'The position values must be numeric.'

    def test_rotation_error(self):
        """Mark image rotation error."""

        for r in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_image()
            assert str(error.exception) == f'The rotation: {r} is not correct. Must be "x y z".'

        for r in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_image()
            assert str(error.exception) == 'The rotation values must be numeric.'

    def test_width_error(self):
        """Mark image width error."""

        for w in NOT_GREATER_THAN_0_MARK_IMAGE_WIDTHS_HEIGHTS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_image(width=w)
            assert str(error.exception) == 'The width must be greater than 0.'

    def test_height_error(self):
        """Mark image height error."""

        for h in NOT_GREATER_THAN_0_MARK_IMAGE_WIDTHS_HEIGHTS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_image(height=h)
            assert str(error.exception) == 'The height must be greater than 0.'



"""AframeXR utils"""

from aframexr.utils.axis_html_creator import *
from aframexr.utils.constants import *
from aframexr.utils.entities_html_creator import *
from aframexr.utils.entity_creator import *
from aframexr.utils.scene_creator import *
from aframexr.utils.validators import *



"""AframeXR aggregate."""

from pandas import DataFrame

from aframexr.utils.validators import AframeXRValidator


class AggregatedFieldDef:
    """Aggregated field definition."""

    def __init__(self, op: str, field: str, as_field: str = ''):
        AframeXRValidator.validate_aggregate_operation(op)
        self.op = op

        if op != 'count' and not field:  # Field must be filled except using "count" operation
            raise ValueError(f'Parameter "field" cannot be empty using {op}.')
        self.field = field

        if not as_field:
            as_field = field
        self.as_field = as_field

    # Import
    @staticmethod
    def from_dict(aggregate_specs: dict):
        """Creates an AggregatedFieldDef object from the aggregate specifications."""

        AframeXRValidator.validate_type(aggregate_specs, dict)

        try:  # Validate that 'field' and 'aggregate' are ono the specifications
            aggregate_op = aggregate_specs['op']
            field = aggregate_specs['field']
            as_field = aggregate_specs.get('as', field)
        except KeyError:
            raise KeyError('Invalid aggregate specification, must contain "op" and "field".')
        return AggregatedFieldDef(aggregate_op, field, as_field)

    # Export
    def to_dict(self) -> dict:
        """Returns the dictionary representation for chart specifications of the aggregated field."""

        specs: dict = {'op': self.op, 'field': self.field}
        if self.as_field:
            specs['as'] = self.as_field
        return specs

    # Utils
    def get_aggregated_data(self, data: DataFrame, groupby: list) -> DataFrame:
        """Returns the aggregated data."""

        try:
            agg_map = {self.as_field: (self.field, self.op)}
            aggregated_data = data.groupby(groupby).agg(**agg_map).reset_index()
        except KeyError as e:
            unknown_fields = e.args[0]  # Fields that are not in data
            raise KeyError(f'Data has no key {unknown_fields}.')
        return aggregated_data

    @staticmethod
    def split_operator_field(aggregate_formula: str):
        """Returns the aggregate operator, the field and the groupby in the aggregate formula."""

        field, aggregate_op = aggregate_formula, None

        if '(' in aggregate_formula and  ')' in aggregate_formula:
            aggregate_op = aggregate_formula.split('(')[0].strip()  # Value before parentheses (aggregate operation)
            AframeXRValidator.validate_aggregate_operation(aggregate_op)  # Validate that the aggregate is correct

            field = aggregate_formula.split('(')[1].split(')')[0].strip()  # Value between parentheses (field)
        return field, aggregate_op
